"""
Luna-admin client module
"""
from numbers import Real
from typing import Awaitable, List, Optional, Union

import aiohttp

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..common.base_api import BaseHTTPAPI
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload
from ..common.utils import createBasicAuth
from ..tasks.http_objs import AEEventFilters, AEFaceFilters
from ..tasks.utils import getTaskPayload, getTaskSchedulePayload


class AdminApi(BaseHTTPAPI):
    """
    Class for request to luna-admin.

    Attributes:
        basicAuth (str): basic auth header

    """

    def __init__(
        self,
        login: str,
        password: str,
        origin: Optional[str] = "http://127.0.0.1:5010",
        api: Optional[int] = 4,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )
        self.basicAuth = createBasicAuth(login, password)

    def makeAuthRequest(self, **requestKwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Prepare auth header and make request.

        Args:
            requestKwargs: standard request kwargs

        Returns:
            structure with status code, request and decoded Luna API response body is returned.

        """
        requestKwargs.setdefault("headers", {}).update(self.basicAuth)
        return self.makeRequest(**requestKwargs)

    def updateSettings(self, login: str, password: str, **kwargs):  # pylint: disable=W0221
        """
        Update settings.

        Args:
            login: new login
            password: new password

        Keyword Args:
            see Attributes
        """
        super().updateSettings(**kwargs)
        self.basicAuth = createBasicAuth(login, password)

    def getSysInfo(self, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get system info.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with system info.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/luna_sys_info", method="GET", **kwargs)

    def search(self, searchParam: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Search account, face, list, attribute by id or email.

        Args:
            searchParam: search param - account/face/list/attribute id (uuid4) or email

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with found element type, element info and basic stats.
        """
        queries = {"q": searchParam}
        return self.makeAuthRequest(url=f"{self.baseUri}/search", method="GET", queryParams=queries, **kwargs)

    def getAccounts(
        self, page: Optional[int] = 1, pageSize: Optional[int] = 10, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get accounts with pagination.

        Args:
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list of account's info and account count.

        """
        queries = {"page": page, "page_size": pageSize}
        return self.makeAuthRequest(url=f"{self.baseUri}/accounts", method="GET", queryParams=queries, **kwargs)

    def getAccount(self, accountId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get account by id.

        Args:
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with account info and basic stats.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/accounts/{accountId}", method="GET", **kwargs)

    def createAccount(
        self, email: str, organizationName: str, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Register an account.

        Args:
            email: email address
            organizationName: organization name

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with account id.

        """
        payload = RequestPayload.buildJson(body={"email": email, "organization_name": organizationName})
        return self.makeAuthRequest(url=f"{self.baseUri}/accounts", method="POST", body=payload, **kwargs)

    def putAccount(
        self, accountId: str, email: str, organizationName: str, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Register account by id.

        Args:
            accountId: account id
            email: email address
            organizationName: organization name

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with account id.

        """
        payload = RequestPayload.buildJson(body={"email": email, "organization_name": organizationName})
        return self.makeAuthRequest(url=f"{self.baseUri}/accounts/{accountId}", method="PUT", body=payload, **kwargs)

    def removeAccount(self, accountId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Remove account by id.

        Args:
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        return self.makeAuthRequest(url=f"{self.baseUri}/accounts/{accountId}", method="DELETE", **kwargs)

    def getLists(
        self, accountId=None, page: Optional[int] = 1, pageSize: Optional[int] = 10, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get lists with pagination.

        Args:
            accountId: account id
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list of list's info and list count.
        """
        queries = {"page": page, "page_size": pageSize, "account_id": accountId}
        return self.makeAuthRequest(url=f"{self.baseUri}/lists", method="GET", queryParams=queries, **kwargs)

    def getList(self, listId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get list by id.

        Args:
            listId: list id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list info.

        """
        return self.makeAuthRequest(url=f"{self.baseUri}/lists/{listId}", method="GET", **kwargs)

    def getTasks(
        self,
        taskIds: Optional[List[int]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        taskStatus: Optional[int] = None,
        scheduleId: Optional[int] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tasks with filters.

        Args:
            taskIds: task ids
            createTimeLt: task create time upper excluding bound
            createTimeGte: task create time lower including bound
            endTimeLt: task end time upper excluding bound
            endTimeGte: task end time lower including bound
            taskType: task type filter
            taskStatus: task status filter
            scheduleId: schedule id filter
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list of task's info.
        """
        queries = {
            "task_ids": taskIds,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "end_time__lt": endTimeLt,
            "end_time__gte": endTimeGte,
            "task_type": taskType,
            "task_status": taskStatus,
            "schedule_id": scheduleId,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks", method="GET", queryParams=queries, **kwargs)

    def getTaskCount(
        self,
        taskIds: Optional[List[int]] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        taskStatus: Optional[int] = None,
        scheduleId: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task count with filters.

        Args:
            taskIds: task ids
            createTimeLt: task create time upper excluding bound
            createTimeGte: task create time lower including bound
            endTimeLt: task end time upper excluding bound
            endTimeGte: task end time lower including bound
            taskType: task type filter
            taskStatus: task status filter
            scheduleId: schedule id filter

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task count.
        """
        queries = {
            "task_ids": taskIds,
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "end_time__lt": endTimeLt,
            "end_time__gte": endTimeGte,
            "task_type": taskType,
            "task_status": taskStatus,
            "schedule_id": scheduleId,
        }
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/count", method="GET", queryParams=queries, **kwargs)

    def getTask(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task by id.

        Args:
            taskId: task id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task info.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/{taskId}", method="GET", **kwargs)

    def cancelTask(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Cancel task by id.

        Args:
            taskId: task id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/{taskId}", method="PATCH", **kwargs)

    def deleteTask(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete task by id.

        Args:
            taskId: task id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/{taskId}", method="DELETE", **kwargs)

    def getTaskResult(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task result by id.

        Args:
            taskId: task id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task result.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/{taskId}/result", method="GET", **kwargs)

    def getTaskErrors(
        self, taskId: int, page: Optional[int] = 1, pageSize: Optional[int] = 10, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task errors by task id.

        Args:
            taskId: task id
            page:  page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list of task error's info.
        """
        queries = {"page": page, "page_size": pageSize}
        return self.makeAuthRequest(
            url=f"{self.baseUri}/tasks/{taskId}/errors", method="GET", queryParams=queries, **kwargs
        )

    def getTaskErrorCount(
        self,
        taskId: int,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        errorCode: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get task errors by task id.

        Args:
            taskId: task id
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            errorCode: task status filter

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.
        """
        queries = {"error_time__lt": errorTimeLt, "error_time__gte": errorTimeGte, "error_code": errorCode}
        return self.makeAuthRequest(
            url=f"{self.baseUri}/tasks/{taskId}/errors/count", method="GET", queryParams=queries, **kwargs
        )

    def getTaskSubtasks(self, taskId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get subtasks by task id.

        Args:
            taskId: task id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list of subtask's info.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/{taskId}/subtasks", method="GET", **kwargs)

    def getErrors(
        self,
        errorIds: Optional[List[int]] = None,
        taskIds: Optional[List[int]] = None,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        errorCode: Optional[int] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get errors with filters.

        Args:
            errorIds: error ids
            taskIds: task ids
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            taskType: task type filter
            errorCode: task status filter
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with list of error's info.
        """
        query = {
            "task_ids": taskIds,
            "error_ids": errorIds,
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "task_type": taskType,
            "error_code": errorCode,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/errors", method="GET", queryParams=query, **kwargs)

    def getErrorCount(
        self,
        errorIds: Optional[List[int]] = None,
        taskIds: Optional[List[int]] = None,
        errorTimeLt: Optional[str] = None,
        errorTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        errorCode: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get error count with filters.

        Args:
            errorIds: error ids
            taskIds: task ids
            errorTimeLt: error time upper excluding bound
            errorTimeGte: error time lower including bound
            taskType: task type filter
            errorCode: task status filter

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with error count.
        """
        query = {
            "task_ids": taskIds,
            "error_ids": errorIds,
            "error_time__lt": errorTimeLt,
            "error_time__gte": errorTimeGte,
            "task_type": taskType,
            "error_code": errorCode,
        }
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/errors/count", method="GET", queryParams=query, **kwargs)

    def getError(self, errorId: int, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get error by id.

        Args:
            errorId: error id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with error info.
        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/errors/{errorId}", method="GET", **kwargs)

    def taskGc(
        self,
        target: str,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        insertTimeLt: Optional[str] = None,
        handlerId: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        descriptorVersion: Optional[int] = None,
        descriptorType: Optional[str] = None,
        gcAccountId: Optional[str] = None,
        description: Optional[str] = "",
        storeResults: Optional[bool] = None,
        removeSamples: Optional[bool] = None,
        removeImageOrigins: Optional[bool] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start gc task

        Args:
            target: garbage collection target:
                   "face_descriptors" to delete descriptors by version from luna-faces
                   "event_descriptors" to delete descriptors by version and type from luna-events
                   "events" to delete events
            descriptorVersion: descriptor version
            descriptorType: descriptor type, "face" or "body"
            description: description
            createTimeLt: event create time upper excluding bound
            createTimeGte: face create time bottom excluding bound
            insertTimeLt: event insert time upper excluding bound
            userData: user data for faces filtering
            listId: list id for faces filtering
            handlerId: event handler id
            gcAccountId: account id to remove entities within
            storeResults: whether to store task results
            removeSamples: whether to remove samples along with the event or face
            removeImageOrigins: whether to remove image origins along with the event

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.
        """
        payload = getTaskPayload(
            taskType="gc",
            description=description,
            target=target,
            createTimeLt=createTimeLt,
            createTimeGte=createTimeGte,
            insertTimeLt=insertTimeLt,
            handlerId=handlerId,
            descriptorVersion=descriptorVersion,
            descriptorType=descriptorType,
            userData=userData,
            listId=listId,
            gcAccountId=gcAccountId,
            storeResults=storeResults,
            removeSamples=removeSamples,
            removeImageOrigins=removeImageOrigins,
        )
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/gc", method="POST", body=payload, **kwargs)

    def taskAdditionalExtract(
        self,
        extractionTarget: str,
        description: str = "",
        missing: bool = True,
        descriptorVersion: Optional[int] = None,
        filters: Union[AEFaceFilters, AEEventFilters] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create and start additional extract task

        Args:
            description: task description
            extractionTarget: what part of an attribute to extract
            missing: whether extract only missing attribute parts
            descriptorVersion: descriptor version to extract | use if 'extractionTarget' is "descriptor"
            filters: faces filters

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with task id.
        """
        taskContent = {
            "description": description,
            "content": {
                "extraction_target": extractionTarget,
                "options": {"missing": missing},
                "filters": filters.asDict(),
            },
        }
        if descriptorVersion is not None:
            taskContent["content"]["options"]["descriptor_version"] = descriptorVersion

        payload = RequestPayload.buildJson(body=taskContent)
        return self.makeAuthRequest(
            url=f"{self.baseUri}/tasks/additional_extract", method="POST", body=payload, **kwargs
        )

    def createSchedule(
        self,
        taskType: Literal[4],
        contentKwargs: dict,
        cron: str,
        cronTz: Optional[str] = None,
        startImmediately: Optional[bool] = None,
        createStopped: Optional[bool] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create schedule

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createSchedule  # pylint: disable=line-too-long

        Args:
            taskType: type of task
            contentKwargs: kwargs for create content
            cron: cron string
            cronTz: cron timezone
            startImmediately: start task immediately
            createStopped: create task stopped schedule task

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with schedule id.

        """
        payload = getTaskSchedulePayload(
            taskType=taskType,
            cron=cron,
            cronTz=cronTz,
            startImmediately=startImmediately,
            createStopped=createStopped,
            **contentKwargs,
        )
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/schedules", method="POST", body=payload, **kwargs)

    def putSchedule(
        self,
        scheduleId: int,
        taskType: int,
        contentKwargs: dict,
        cron: str,
        cronTz: Optional[str] = None,
        startImmediately: Optional[bool] = None,
        createStopped: Optional[bool] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update schedule

        Args:
            scheduleId: schedule id
            taskType: type of task
            contentKwargs: kwargs for create content
            cron: cron string
            cronTz: cron timezone
            startImmediately: start task immediately
            createStopped: create task stopped schedule task

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with schedule id.

        """
        payload = getTaskSchedulePayload(
            taskType=taskType,
            cron=cron,
            cronTz=cronTz,
            startImmediately=startImmediately,
            createStopped=createStopped,
            **contentKwargs,
        )
        return self.makeAuthRequest(
            url=f"{self.baseUri}/tasks/schedules/{scheduleId}", method="PUT", body=payload, **kwargs
        )

    def patchSchedule(
        self,
        scheduleId: int,
        action: str,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Patch schedule by id

        Args:
            scheduleId: schedule id
            action: action for schedule

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {"action": action}
        return self.makeAuthRequest(
            url=f"{self.baseUri}/tasks/schedules/{scheduleId}", method="PATCH", queryParams=query, **kwargs
        )

    def getSchedules(
        self,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        taskType: Optional[int] = None,
        page: Optional[int] = None,
        pageSize: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get schedules with filters

        Args:
            createTimeLt: schedule create time upper excluding bound
            createTimeGte: schedule create time lower including bound
            taskType: task type filter
            accountIdFilter: account id to get schedules
            page: page
            pageSize: page size

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        query = {
            "create_time__lt": createTimeLt,
            "create_time__gte": createTimeGte,
            "task_type": taskType,
            "page": page,
            "page_size": pageSize,
        }
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/schedules/", method="GET", queryParams=query, **kwargs)

    def getSchedule(
        self,
        scheduleId: int,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get schedule by id

        Args:
            scheduleId: schedule id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/schedules/{scheduleId}", method="GET", **kwargs)

    def deleteSchedule(
        self,
        scheduleId: int,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete schedule by id

        Args:
            scheduleId: schedule id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with response.

        """
        return self.makeAuthRequest(url=f"{self.baseUri}/tasks/schedules/{scheduleId}", method="DELETE", **kwargs)
