"""
Module implements classes for standard structures.
"""
# pylint: disable=C0103, R0914

import base64
from dataclasses import dataclass
from numbers import Real
from typing import Any, Dict, Iterable, List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from luna3.common.luna_request import guessContentType
from luna3.events.http_objs import CLOTHING_COLOR, PRIMITIVE_CLOTHING_COLOR

MISSING = object()

BASE64_IMAGE_TYPES_MAP = {
    "image/jpeg": "image/x-jpeg-base64",
    "image/png": "image/x-png-base64",
    "image/bmp": "image/x-bmp-base64",
    "image/tiff": "image/x-tiff-base64",
    "image/x-portable-pixmap": "image/x-portable-pixmap-base64",
}


class BoundingBox:
    """
    Face bounding box

    Attributes:
        width (int): width
        height (int): height
        x (int): x coordinate of top left angle
        y (int): y coordinate of top left angle
    """

    __slots__ = ("x", "y", "width", "height")

    def __init__(self, width: int, height: int, x: int, y: int):
        self.x: int = x  # pylint: disable=C0103
        self.y: int = y  # pylint: disable=C0103
        self.width: int = width
        self.height: int = height

    def asDict(self):
        """
        Get bounding box as dict
        Returns:
            dict with bounding box
        """
        return {"x": self.x, "y": self.y, "width": self.width, "height": self.height}


class UrlForDetection:
    """
    Url for detection class

    Attributes:
        url: url for detection
        faceBoundingBoxList (Optional[List[BoundingBox]]): face bounding box list
        bodyBoundingBoxList (Optional[List[BoundingBox]]): human body bounding box list
        detectTime (Optional[str]): image detection time in ISO format
        detectTs (Optional[float]): user-defined timestamp relative to something, such as the start of a video
        imageOrigin (Optional[str]): image detection origin
        imageOriginMimetype (Optional[str]): image origin mimetype, used when imageOrigin is base64-encoded image
        meta (Optional[dict]): user-defined event metadata
    """

    __slots__ = (
        "url",
        "faceBoundingBoxList",
        "bodyBoundingBoxList",
        "detectTime",
        "detectTs",
        "imageOrigin",
        "imageOriginMimetype",
        "meta",
    )

    def __init__(
        self,
        url: str,
        faceBoundingBoxList: Optional[List[BoundingBox]] = None,
        bodyBoundingBoxList: Optional[List[BoundingBox]] = None,
        detectTime: Optional[str] = None,
        detectTs: Optional[float] = None,
        imageOrigin: Optional[str] = None,
        imageOriginMimetype: Optional[str] = None,
        meta: Optional[dict] = None,
    ) -> None:
        self.url = url
        self.faceBoundingBoxList = faceBoundingBoxList
        self.bodyBoundingBoxList = bodyBoundingBoxList
        self.detectTime = detectTime
        self.detectTs = detectTs
        self.imageOrigin = imageOrigin
        self.imageOriginMimetype = imageOriginMimetype
        self.meta = meta

    def asDict(self):
        """
        Get url for detection as dict
        Returns:
            dict with url for detection and bounding box
        """
        res = {"url": self.url}
        if self.faceBoundingBoxList is not None:
            res["face_bounding_boxes"] = [boundingBox.asDict() for boundingBox in self.faceBoundingBoxList]
        if self.bodyBoundingBoxList is not None:
            res["body_bounding_boxes"] = [boundingBox.asDict() for boundingBox in self.bodyBoundingBoxList]
        if self.detectTime is not None:
            res["detect_time"] = self.detectTime
        if self.detectTs is not None:
            res["detect_ts"] = self.detectTs
        if self.imageOrigin is not None:
            res["image_origin"] = self.imageOrigin
        if self.imageOriginMimetype is not None:
            res["image_origin_mimetype"] = self.imageOriginMimetype
        if self.meta is not None:
            res["meta"] = self.meta
        return res


class SampleForDetection:
    """
    Sample for detection class

    Attributes:
        sampleId (str): sample id
        detectTime (Optional[str]): sample detection time in ISO format
        imageOrigin (Optional[str]): sample detection image origin
        imageOriginMimetype (Optional[str]): image origin mimetype, used when imageOrigin is base64-encoded image
        detectTs (Optional[float]): user-defined timestamp relative to something, such as the start of a video
        meta (Optional[dict]): user-defined event metadata
    """

    __slots__ = ("sampleId", "detectTime", "detectTs", "imageOrigin", "imageOriginMimetype", "meta")

    def __init__(
        self,
        sampleId: str,
        detectTime: Optional[str] = None,
        detectTs: Optional[float] = None,
        imageOrigin: Optional[str] = None,
        imageOriginMimetype: Optional[str] = None,
        meta: Optional[dict] = None,
    ) -> None:
        self.sampleId = sampleId
        self.detectTime = detectTime
        self.imageOrigin = imageOrigin
        self.imageOriginMimetype = imageOriginMimetype
        self.detectTs = detectTs
        self.meta = meta

    def asDict(self):
        """
        Get sample for detection as dict
        Returns:
            dict with sample id and time of its detection
        """
        res = {"sample_id": self.sampleId}
        if self.detectTime is not None:
            res["detect_time"] = self.detectTime
        if self.detectTs is not None:
            res["detect_ts"] = self.detectTs
        if self.imageOrigin is not None:
            res["image_origin"] = self.imageOrigin
        if self.imageOriginMimetype is not None:
            res["image_origin_mimetype"] = self.imageOriginMimetype
        if self.meta is not None:
            res["meta"] = self.meta
        return res


class BinaryData:
    """
    Binary Data class.

    Attributes:
        filename: filename.
        path: filename. may be path to the image.
        mimetype: property. image mimetype. Can be autodetected from body.
        _mimetype: private store mimetype
        body: property. binary representation of image. Can be uploaded by filename path.
        _body: private store body bytes
        multipartHeaders: Custom headers for multipart body

    """

    # Available mimetypes (should be defied in a derived class)
    availableMimetypes = tuple()

    def __init__(
        self,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        mimetype: str = None,
        body: bytes = None,
        multipartHeaders: dict = None,
    ) -> None:
        """
        Init

        Args:
            path: path for load image from file
            filename: file name (is used for multipart ContentDisposition header)
            mimetype: mimetype
            body: binary body
            multipartHeaders: multipartHeaders

        """
        self.path = path
        self.filename = filename if filename is not None else path.split("/")[-1]
        self._body = body
        self._mimetype = mimetype

        if multipartHeaders is None:
            self.multipartHeaders = {}
        else:
            self.multipartHeaders = multipartHeaders

    @property
    def body(self) -> bytes:
        """
        Lazy body. If not set, uploads itself by the filename path.

        Returns:
             body
        """
        if self._body is not None:
            return self._body

        with open(self.path, "rb") as file:
            self._body = file.read()
        return self._body

    @body.setter
    def body(self, body: bytes):
        """
        Set body.

        Args:
            body: image bytes
        """
        self._body = body

    @property
    def mimetype(self) -> str:
        """
        Lazy mimetype. If not set, executed from body.

        Returns:
             mimetype
        """
        if self._mimetype is not None:
            return self._mimetype
        self._mimetype = self.getMimetype(self.body)
        return self._mimetype

    @mimetype.setter
    def mimetype(self, mimetype: str) -> None:
        """
        Set mimetype

        Args:
            image mimetype
        """
        self._mimetype = mimetype

    @property
    def multipartContentDispositionHeader(self) -> str:
        """
        Generate ContentDisposition header.

        Returns:
             Content-Disposition
        """
        return f'form-data; filename="{self.filename}"'

    @property
    def httpMultipartHeaders(self) -> Dict[str, str]:
        """
        Generate headers for each part of the multipart body.

        Returns:
             Content-Disposition and Content-Type headers.
        """
        return {
            "Content-Disposition": self.multipartContentDispositionHeader,
            "Content-Type": self.mimetype,
            **self.multipartHeaders,
        }

    def getMimetype(self, body: bytes) -> str:
        """ "
        The function gets mimetype of body.

        Args:
            body: binary data

        Raises:
            ValueError: if body has wrong mimetype

        Returns:
             Content-type of the body.
        """
        mimetype = guessContentType(body)
        if mimetype not in self.availableMimetypes:
            raise ValueError(f"Unsupported mimetype, availbale mimetypes are: {self.availableMimetypes}")
        return mimetype


class BinaryImageOrigin(BinaryData):
    """Binary Image Origin class."""

    availableMimetypes = ("image/jpeg", "image/png", "image/gif", "image/bmp", "image/tiff", "image/x-portable-pixmap")


class BinaryImage(BinaryData):
    """
    Binary Image class.

    Attributes:
        filename: filename.
        path: filename. may be path to the image.
        mimetype: property. image mimetype. Can be autodetected from body.
        _mimetype: private store mimetype
        body: property. binary representation of image. Can be uploaded by filename path.
        _body: private store body bytes
        multipartHeaders: Custom headers for multipart body
        detectTime (Optional[str]): image detection time in ISO format
        detectTs (Optional[float]): user-defined timestamp relative to something, such as the start of a video
        faceBoundingBoxList (Optional[List[BoundingBox]]): list of faces bounding boxes
        bodyBoundingBoxList (Optional[List[BoundingBox]]): list of human bodies bounding boxes
        imageOrigin (Optional[str]): image origin
        imageOriginMimetype (Optional[str]): image origin mimetype, used when imageOrigin is base64-encoded image
        meta (Optional[dict]): user-defined event metadata
    """

    availableMimetypes = ("image/jpeg", "image/png", "image/gif", "image/bmp", "image/tiff", "image/x-portable-pixmap")

    def __init__(
        self,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        mimetype: str = None,
        body: bytes = None,
        multipartHeaders: dict = None,
        detectTime: Optional[str] = None,
        detectTs: Optional[float] = None,
        faceBoundingBoxList: Optional[List[BoundingBox]] = None,
        bodyBoundingBoxList: Optional[List[BoundingBox]] = None,
        imageOrigin: Optional[Union[str, BinaryImageOrigin]] = None,
        imageOriginMimetype: Optional[str] = None,
        meta: Optional[dict] = None,
    ) -> None:
        """
        Init

        Args:
            path: path for load image from file
            filename: file name (is used for multipart ContentDisposition header)
            mimetype: mimetype
            body: binary body
            multipartHeaders: multipartHeaders
            detectTime: image detection time in ISO format
            detectTs: user-defined timestamp relative to something, such as the start of a video
            faceBoundingBoxList: list of faces bounding boxes
            bodyBoundingBoxList: list of bodies bounding boxes
            imageOrigin: image origin
            imageOriginMimetype: image origin mimetype, used when imageOrigin is base64-encoded image
            meta: user-defined event metadata
        """
        super().__init__(path=path, filename=filename, mimetype=mimetype, body=body, multipartHeaders=multipartHeaders)

        self.faceBoundingBoxList = faceBoundingBoxList
        self.bodyBoundingBoxList = bodyBoundingBoxList
        self.detectTime = detectTime
        self.imageOrigin = imageOrigin
        self.imageOriginMimetype = imageOriginMimetype
        self.detectTs = detectTs
        self.meta = meta

    def clone(self) -> "BinaryImage":
        """Clone instance"""
        return BinaryImage(
            filename=self.filename,
            path=self.path,
            mimetype=self._mimetype,
            body=self._body,
            multipartHeaders=self.multipartHeaders,
            detectTime=self.detectTime,
            detectTs=self.detectTs,
            faceBoundingBoxList=self.faceBoundingBoxList,
            bodyBoundingBoxList=self.bodyBoundingBoxList,
            imageOrigin=self.imageOrigin,
            imageOriginMimetype=self.imageOriginMimetype,
            meta=self.meta,
        )

    def asJson(self) -> dict:
        """
        Get Binary image as json with body decoded to str
        Returns:
            dict with body, mimetype and bounding box list if exists
        """
        if self.mimetype.endswith("base64"):
            image = self.body.decode()
            mimetype = self.mimetype
        else:
            image = base64.b64encode(self.body).decode()
            mimetype = BASE64_IMAGE_TYPES_MAP[self.mimetype]
        json = {"image": image, "mimetype": mimetype}
        if self.faceBoundingBoxList is not None:
            json["face_bounding_boxes"] = [boundingBox.asDict() for boundingBox in self.faceBoundingBoxList]
        if self.bodyBoundingBoxList is not None:
            json["body_bounding_boxes"] = [boundingBox.asDict() for boundingBox in self.bodyBoundingBoxList]
        if self.detectTime is not None:
            json["detect_time"] = self.detectTime
        if self.detectTs is not None:
            json["detect_ts"] = self.detectTs
        if self.imageOrigin is not None:
            json["image_origin"] = self.imageOrigin
        if self.imageOriginMimetype is not None:
            json["image_origin_mimetype"] = self.imageOriginMimetype
        if self.meta is not None:
            json["meta"] = self.meta
        return json


class BinaryDescriptorData(BinaryData):
    """Binary Descriptor Data class."""

    availableMimetypes = ("application/x-sdk-descriptor", "application/x-vl-xpk")


class StatQuery:
    """
    Object for easy create query to event service for statistic
    For detail see Luna Events User Manual on *POST* */events/stats*
    """

    TARGETS = [
        "age",
        "apparent_age",
        "apparent_gender",
        "area",
        "backpack_state",
        "city",
        "create_time",
        "district",
        "emotion",
        "end_time",
        "ethnic_group",
        "event_id",
        "external_id",
        "face_id",
        "gender",
        "geo_position",
        "handler_id",
        "headwear_state",
        "house_number",
        "liveness",
        "mask",
        "sleeve_length",
        "source",
        "stream_id",
        "street",
        "tag",
        "top_matching_candidates_label",
        "top_similar_external_id",
        "top_similar_object_id",
        "top_similar_object_similarity",
        "top_similar_object_type",
        "track_id",
        "upper_clothing_colors",
        "user_data",
        "lower_garment_type",
        "shoes_apparent_color",
        "lower_garment_colors",
        "headwear_apparent_color",
    ]

    FILTERS = [
        "account_id",
        "age",
        "apparent_age",
        "apparent_gender",
        "area",
        "backpack_state",
        "city",
        "district",
        "emotion",
        "ethnic_group",
        "event_id",
        "external_id",
        "face_id",
        "gender",
        "geo_position",
        "handler_id",
        "headwear_state",
        "house_number",
        "liveness",
        "mask",
        "sleeve_length",
        "source",
        "stream_id",
        "street",
        "tag",
        "top_matching_candidates_label",
        "top_similar_external_id",
        "top_similar_object_id",
        "top_similar_object_similarity",
        "top_similar_object_type",
        "track_id",
        "upper_clothing_colors",
        "user_data",
        "lower_garment_type",
        "shoes_apparent_color",
        "lower_garment_colors",
        "headwear_apparent_color",
    ]

    def __init__(self):
        self.query = {}

    def addTarget(self, column: Literal[TARGETS], aggregator: str = "count") -> "StatQuery":
        """
        Add target query field

        Args:
            column: column for select
            aggregator: aggregation function (count, max, min, avg, group_by)

        Returns:
            self
        """
        self.query.setdefault("targets", []).append({"column": column, "aggregator": aggregator})
        return self

    def addFilter(
        self, column: Literal[FILTERS], operator: str, value: Union[str, list, float, int, None]
    ) -> "StatQuery":
        """
        Add filter query field

        For column 'tags' work only operator 'like/nlike' on each tag.

        Args:
            column: column for filter
            operator: filter operation (eq, neq, in, nin, like, nlike, gt, gte, lt, lte).
                      For tags must be like/nlike only.
            value: filter value

        Returns:
            self
        """
        self.query.setdefault("filters", []).append({"column": column, "operator": operator, "value": value})
        return self

    def addPeriod(self, operator: str, value: str) -> "StatQuery":
        """
        Add period query field

        Args:
            operator: period operation (gt, gte, lt, lte)
            value: value for operation (time in ISO or now-time format)

        Returns:
            self
        """
        self.query.setdefault("period", []).append({"operator": operator, "value": value})
        return self

    def addGroupBy(self, groupBy: str) -> "StatQuery":
        """
        Add group by query field

        Args:
            groupBy: events histogram "group by" time clause in frequency grouping: minuteOfHour, hourOfDay,
                dayOfWeek, dayOfMonth, dayOfYear, weekOfYear, monthOfYear
                or periodic grouping (y|M|w|d|h|m|s)+int

        Returns:
            self
        """
        self.query["group_by"] = groupBy
        return self


def getNonEmptyFilters(filters: dict) -> dict:
    """
    Returns all filters that have not None value
    Args:
        filters: filters

    Returns:
        dict with non empty filters
    """
    resFilters = {key: value for key, value in filters.items() if value is not None}
    return resFilters


class _ObjectFilters:
    """
    Root Object filters class.

    Class Attributes:
        objectsType: current objects type
    """

    # empty object type
    objectsType: str = None

    def __init__(self):
        self.filters = {}

    @property
    def values(self) -> dict:
        """
        Get nonempty values from self.filters.

        Returns:
            nonempty values from self.filters
        """
        return getNonEmptyFilters(self.filters)

    def __repr__(self):
        return f"<{self.__class__.__name__}>: {self.values}"


class RawDescriptor:
    """
    Raw descriptor.
    Attributes:
        version (int): descriptor version
        descriptor (bytes): descriptor
    """

    __slots__ = ("version", "descriptor")

    def __init__(self, version: int, descriptor: bytes):
        self.version = version
        self.descriptor = descriptor

    def __eq__(self, other):
        if not isinstance(other, RawDescriptor):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.version == other.version and self.descriptor == other.descriptor

    def __hash__(self):
        return hash((self.version, self.descriptor))


class IntThreshold:
    """
    Threshold with min and max values of int format

    Attributes:
        minimum: minimum
        maximum: maximum
    """

    def __init__(self, minimum: Optional[int] = None, maximum: Optional[int] = None):
        self.minimum = minimum
        self.maximum = maximum

    def asDict(self) -> Dict[str, int]:
        """Created dictionary representation"""
        result = {}
        if self.minimum is not None:
            result["min"] = self.minimum
        if self.maximum is not None:
            result["max"] = self.maximum
        return result


class FloatThreshold:
    """
    Threshold with min and max values of float format

    Attributes:
        minimum: minimum
        maximum: maximum
    """

    def __init__(self, minimum: Optional[float] = None, maximum: Optional[float] = None):
        self.minimum = minimum
        self.maximum = maximum

    def asDict(self) -> Dict[str, float]:
        """Created dictionary representation"""
        result = {}
        if self.minimum is not None:
            result["min"] = self.minimum
        if self.maximum is not None:
            result["max"] = self.maximum
        return result


class ImageFormatCheck:
    """
    Image format check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[List[Literal["JPEG", "JPEG2000", "PNG"]]] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


class FloatQualityCheck:
    """
    Face quality check structure for check with float threshold
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[FloatThreshold] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold.asDict()
        return result


class IntQualityCheck:
    """
    Face quality check structure for check with int threshold
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[IntThreshold] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold.asDict()
        return result


class GlassesCheck:
    """
    Glasses check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(
        self, estimate: bool = True, threshold: Optional[List[Literal["no_glasses", "eyeglasses", "sunglasses"]]] = None
    ):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


class LeftEyeCheck:
    """
    Left eye check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[List[Literal["open", "occluded", "closed"]]] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


class RightEyeCheck:
    """
    Right eye check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    def __init__(self, estimate: bool = True, threshold: Optional[List[Literal["open", "occluded", "closed"]]] = None):
        self.estimate = estimate
        self.threshold = threshold

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = list(self.threshold)
        return result


@dataclass
class Check:
    """
    Base check structure
    Attributes:
        estimate: whether to enable check
        threshold: threshold
    """

    estimate: bool = True
    threshold: Optional[Any] = None

    def asDict(self) -> Dict[str, Any]:
        """Created dictionary representation"""
        result = {"estimate": int(self.estimate)}
        if self.threshold is not None:
            result["threshold"] = self.threshold
        return result


@dataclass
class EyebrowsCheck(Check):
    """Check for eyebrows_type."""

    threshold: Iterable[Literal["squinting", "frowning", "raised", "neutral"]] = ("neutral",)


@dataclass
class ShouldersPositionCheck(Check):
    """Check for shoulders_position."""

    threshold: Iterable[Literal["non-parallel", "parallel", "hidden"]] = ("parallel",)


@dataclass
class SmileCheck(Check):
    """Check for smile_properties."""

    threshold: Iterable[Literal["none", "smile_lips", "smile_teeth"]] = ("none",)


@dataclass
class HeadwearCheck(Check):
    """Check for headwear_type."""

    threshold: Iterable[
        Literal[
            "none",
            "baseball_cap",
            "beanie",
            "peaked_cap",
            "shawl",
            "hat_with_earflaps",
            "helmet",
            "hood",
            "hat",
            "other",
        ]
    ] = ("none",)


@dataclass
class FaceColorCheck(Check):
    """Check for face_color_type."""

    threshold: Iterable[Literal["color", "grayscale", "infrared"]] = ("color",)


@dataclass
class FlagCheck(Check):
    """Universal check with 1/0 threshold."""

    threshold: Literal[0, 1] = 0


class FaceQualityChecks:
    """
    Face quality checks

    Attributes:
        imageFormat: image format check
        illuminationQuality: illumination quality check
        specularityQuality: specularity quality check
        blurrinessQuality: blurriness quality check
        darkQuality: dark quality check
        lightQuality: light quality check
        headYaw: head yaw check
        headPitch: head pitch check
        headRoll: head roll check
        gazeYaw: gaze yaw check
        gazePitch: gaze pitch check
        mouthSmiling: mouth smiling check
        mouthOccluded: mouth occluded check
        mouthOpen: mouth open check
        glasses: glasses check
        leftEye: left eye check
        rightEye: right eye check
        headHorizontalCenter: head horizontal center check
        headVerticalCenter: head vertical center check
        headWidth: head width check
        headHeight: head height check
        eyeDistance: eye distance check
        imageWidth: image width check
        imageHeight: image height check
        aspectRatio: aspect ratio check
        faceWidth: face width check
        faceHeight: face height check
        indentLeft: indent left check
        indentRight: indent right check
        indentUpper: indent upper check
        indentLower: indent lower check
        imageSize: image size check
        eyebrowsState: check configuration
        smileProperties: check configuration
        headwearType: check configuration
        naturalLight: check configuration
        radialDistortion: check configuration
        redEyes: check configuration
        faceColorType: check configuration
        illuminationUniformity: illumination uniformity check
        dynamicRange: dynamic range check
        backgroundLightness: face detection background lightness
        backgroundUniformity: face detection background uniformity
        shouldersPosition: shoulders position status
    """

    def __init__(
        self,
        imageFormat: Optional[ImageFormatCheck] = None,
        illuminationQuality: Optional[FloatQualityCheck] = None,
        specularityQuality: Optional[FloatQualityCheck] = None,
        blurrinessQuality: Optional[FloatQualityCheck] = None,
        darkQuality: Optional[FloatQualityCheck] = None,
        lightQuality: Optional[FloatQualityCheck] = None,
        headYaw: Optional[FloatQualityCheck] = None,
        headPitch: Optional[FloatQualityCheck] = None,
        headRoll: Optional[FloatQualityCheck] = None,
        gazeYaw: Optional[FloatQualityCheck] = None,
        gazePitch: Optional[FloatQualityCheck] = None,
        mouthSmiling: Optional[FloatQualityCheck] = None,
        mouthOccluded: Optional[FloatQualityCheck] = None,
        mouthOpen: Optional[FloatQualityCheck] = None,
        glasses: Optional[GlassesCheck] = None,
        leftEye: Optional[LeftEyeCheck] = None,
        rightEye: Optional[RightEyeCheck] = None,
        headHorizontalCenter: Optional[FloatQualityCheck] = None,
        headVerticalCenter: Optional[FloatQualityCheck] = None,
        headWidth: Optional[FloatQualityCheck] = None,
        headHeight: Optional[FloatQualityCheck] = None,
        eyeDistance: Optional[FloatQualityCheck] = None,
        imageWidth: Optional[IntQualityCheck] = None,
        imageHeight: Optional[IntQualityCheck] = None,
        aspectRatio: Optional[FloatQualityCheck] = None,
        faceWidth: Optional[IntQualityCheck] = None,
        faceHeight: Optional[IntQualityCheck] = None,
        indentLeft: Optional[IntQualityCheck] = None,
        indentRight: Optional[IntQualityCheck] = None,
        indentLower: Optional[IntQualityCheck] = None,
        indentUpper: Optional[IntQualityCheck] = None,
        imageSize: Optional[IntQualityCheck] = None,
        eyebrowsState: Optional[EyebrowsCheck] = None,
        smileProperties: Optional[SmileCheck] = None,
        headwearType: Optional[HeadwearCheck] = None,
        naturalLight: Optional[FlagCheck] = None,
        radialDistortion: Optional[FlagCheck] = None,
        redEyes: Optional[FlagCheck] = None,
        backgroundLightness: Optional[FloatQualityCheck] = None,
        backgroundUniformity: Optional[FloatQualityCheck] = None,
        faceColorType: Optional[FaceColorCheck] = None,
        illuminationUniformity: Optional[FloatQualityCheck] = None,
        dynamicRange: Optional[FloatQualityCheck] = None,
        shouldersPosition: Optional[ShouldersPositionCheck] = None,
    ):
        self.imageFormat = imageFormat
        self.illuminationQuality = illuminationQuality
        self.specularityQuality = specularityQuality
        self.blurrinessQuality = blurrinessQuality
        self.darkQuality = darkQuality
        self.lightQuality = lightQuality
        self.headYaw = headYaw
        self.headPitch = headPitch
        self.headRoll = headRoll
        self.gazeYaw = gazeYaw
        self.gazePitch = gazePitch
        self.mouthSmiling = mouthSmiling
        self.mouthOccluded = mouthOccluded
        self.mouthOpen = mouthOpen
        self.glasses = glasses
        self.leftEye = leftEye
        self.rightEye = rightEye
        self.headHorizontalCenter = headHorizontalCenter
        self.headVerticalCenter = headVerticalCenter
        self.headWidth = headWidth
        self.headHeight = headHeight
        self.eyeDistance = eyeDistance
        self.imageWidth = imageWidth
        self.imageHeight = imageHeight
        self.aspectRatio = aspectRatio
        self.faceWidth = faceWidth
        self.faceHeight = faceHeight
        self.indentLeft = indentLeft
        self.indentRight = indentRight
        self.indentUpper = indentUpper
        self.indentLower = indentLower
        self.imageSize = imageSize
        self.eyebrowsState = eyebrowsState
        self.smileProperties = smileProperties
        self.headwearType = headwearType
        self.naturalLight = naturalLight
        self.radialDistortion = radialDistortion
        self.redEyes = redEyes
        self.faceColorType = faceColorType
        self.illuminationUniformity = illuminationUniformity
        self.dynamicRange = dynamicRange
        self.backgroundLightness = backgroundLightness
        self.backgroundUniformity = backgroundUniformity
        self.shouldersPosition = shouldersPosition

    def asDict(self) -> Dict[str, Any]:
        """Creates dictionary representation"""
        result = {
            "image_format": self.imageFormat,
            "illumination_quality": self.illuminationQuality,
            "specularity_quality": self.specularityQuality,
            "blurriness_quality": self.blurrinessQuality,
            "dark_quality": self.darkQuality,
            "light_quality": self.lightQuality,
            "head_yaw": self.headYaw,
            "head_pitch": self.headPitch,
            "head_roll": self.headRoll,
            "gaze_yaw": self.gazeYaw,
            "gaze_pitch": self.gazePitch,
            "mouth_smiling": self.mouthSmiling,
            "mouth_occluded": self.mouthOccluded,
            "mouth_open": self.mouthOpen,
            "glasses": self.glasses,
            "left_eye": self.leftEye,
            "right_eye": self.rightEye,
            "head_horizontal_center": self.headHorizontalCenter,
            "head_vertical_center": self.headVerticalCenter,
            "head_width": self.headWidth,
            "head_height": self.headHeight,
            "eye_distance": self.eyeDistance,
            "image_width": self.imageWidth,
            "image_height": self.imageHeight,
            "aspect_ratio": self.aspectRatio,
            "face_width": self.faceWidth,
            "face_height": self.faceHeight,
            "indent_left": self.indentLeft,
            "indent_right": self.indentRight,
            "indent_upper": self.indentUpper,
            "indent_lower": self.indentLower,
            "image_size": self.imageSize,
            "eyebrows_state": self.eyebrowsState,
            "smile_properties": self.smileProperties,
            "headwear_type": self.headwearType,
            "natural_light": self.naturalLight,
            "radial_distortion": self.radialDistortion,
            "red_eyes": self.redEyes,
            "face_color_type": self.faceColorType,
            "illumination_uniformity": self.illuminationUniformity,
            "dynamic_range": self.dynamicRange,
            "background_lightness": self.backgroundLightness,
            "background_uniformity": self.backgroundUniformity,
            "shoulders_position": self.shouldersPosition,
        }
        return {key: value.asDict() for key, value in result.items() if value is not None}


class PeopleCount:
    """
    Class describes people count estimation policy

    Attributes:
        estimate: whether to estimate people count
    """

    def __init__(self, estimate: bool = False):
        self.estimate = estimate

    def asDict(self) -> Dict[str, Any]:
        """Creates dictionary representation"""
        result = {"estimate": int(self.estimate)}
        return result


class FaceQuality:
    """
    Class describes face quality structure

    Attributes:
        estimate: whether to estimate face quality
        enableFilter: whether to enable face quality filtering
        checks: face quality checks
    """

    def __init__(
        self, estimate: bool = False, enableFilter: bool = False, checks: FaceQualityChecks = FaceQualityChecks()
    ):
        self.estimate = estimate
        self.enableFilter = enableFilter
        self.checks = checks

    def asDict(self) -> Dict[str, Any]:
        """Creates dictionary representation"""
        result = {"estimate": int(self.estimate), "filter": int(self.enableFilter)}
        if self.checks is not None:
            result["checks"] = self.checks.asDict()
        return result


class BaseDetectPolicy:
    """
    Class describes detect policy base structure

    Attributes:
        multifacePolicy: multi-face policy:
                         0 - multi-face detection not allowed,
                         1 - multi-face detection allowed,
                         2 - get best detection from the image
        estimateHeadPose: estimated head pose
        estimateEmotions: whether to estimate emotions from the image
        estimateMask: whether to estimate mask from the image
        estimateQuality: estimate head pose
        estimateGaze: whether to estimate eye gaze from the image.
        estimateEyesAttributes: whether to estimate eye attributes from the image
        estimateMouthAttributes: whether to estimate mouth attributes from the image
        detectLandmarks68: whether to detect basic 68-point facial landmarks on the image
        extractExif: whether to extract EXIF meta information from input JPEG images.
                     Exact output will vary since there are no mandatory requirements to both authoring
                     software and digital cameras how to write the data.
                     This function will only parse the tags and output their names and values as is.
                     Please refer to JEITA CP-3451 EXIF specification for details.
        yawThreshold: maximum deviation yaw angle from 0
        rollThreshold: maximum deviation roll angle from 0
        pitchThreshold: maximum deviation pitch angle from 0
        maskStates: Array of integer (filter by mask state) (Items Enum:1 2 3)
        estimateLiveness: liveness estimation policy
        livenessStates: Array of integer (filter by liveness state) (Items Enum:0 1 2)
        faceQuality: face quality
        estimageGlasses: whether to estimate glasses
    """

    class LivenessEstimationPolicy:
        """
        Class describes liveness estimation policy

        Attributes:
            estimate: whether to estimate liveness
            qualityThreshold: liveness estimation quality theshold
            livenessThreshold: liveness estimation score threshold
        """

        def __init__(
            self, estimate: int = 0, qualityThreshold: Optional[float] = None, livenessThreshold: Optional[float] = None
        ):
            self.estimate = estimate
            self.qualityThreshold = qualityThreshold
            self.livenessThreshold = livenessThreshold

        def asDict(self) -> Dict[str, Any]:
            """Creates dictionary representation."""
            allArgs = {"estimate": self.estimate}
            if self.qualityThreshold is not None:
                allArgs["quality_threshold"] = self.qualityThreshold
            if self.livenessThreshold is not None:
                allArgs["liveness_threshold"] = self.livenessThreshold

            return allArgs

    def __init__(
        self,
        multifacePolicy: int = 1,
        estimateHeadPose: int = 0,
        estimateEmotions: int = 0,
        estimateMask: int = 0,
        estimateQuality: int = 0,
        estimateGaze: int = 0,
        estimateEyesAttributes: int = 0,
        estimateMouthAttributes: int = 0,
        detectLandmarks68: int = 0,
        extractExif: int = 0,
        faceQuality: FaceQuality = FaceQuality(),
        yawThreshold: Optional[int] = None,
        rollThreshold: Optional[int] = None,
        pitchThreshold: Optional[int] = None,
        maskStates: Optional[List[int]] = None,
        estimateLiveness: Optional[LivenessEstimationPolicy] = None,
        livenessStates: Optional[List[int]] = None,
        estimateGlasses: int = 0,
    ):
        self.multifacePolicy = multifacePolicy
        self.estimateHeadPose = estimateHeadPose
        self.estimateEmotions = estimateEmotions
        self.estimateMask = estimateMask
        self.estimateQuality = estimateQuality
        self.estimateGaze = estimateGaze
        self.estimateEyesAttributes = estimateEyesAttributes
        self.estimateMouthAttributes = estimateMouthAttributes
        self.detectLandmarks68 = detectLandmarks68
        self.extractExif = extractExif
        self.yawThreshold = yawThreshold
        self.rollThreshold = rollThreshold
        self.pitchThreshold = pitchThreshold
        self.maskStates = maskStates
        self.estimateLiveness = estimateLiveness or DetectPolicy.LivenessEstimationPolicy()
        self.livenessStates = livenessStates
        self.faceQuality = faceQuality
        self.estimateGlasses = estimateGlasses

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        allArgs = {
            "multiface_policy": self.multifacePolicy,
            "estimate_head_pose": self.estimateHeadPose,
            "estimate_emotions": self.estimateEmotions,
            "estimate_liveness": self.estimateLiveness.asDict(),
            "estimate_mask": self.estimateMask,
            "estimate_quality": self.estimateQuality,
            "estimate_gaze": self.estimateGaze,
            "estimate_eyes_attributes": self.estimateEyesAttributes,
            "estimate_mouth_attributes": self.estimateMouthAttributes,
            "detect_landmarks68": self.detectLandmarks68,
            "extract_exif": self.extractExif,
            "estimate_glasses": self.estimateGlasses,
        }
        if self.yawThreshold is not None:
            allArgs["yaw_threshold"] = self.yawThreshold
        if self.rollThreshold is not None:
            allArgs["roll_threshold"] = self.rollThreshold
        if self.pitchThreshold is not None:
            allArgs["pitch_threshold"] = self.pitchThreshold
        if self.maskStates is not None:
            allArgs["mask_states"] = self.maskStates
        if self.livenessStates is not None:
            allArgs["liveness_states"] = self.livenessStates
        if self.faceQuality is not None:
            allArgs["face_quality"] = self.faceQuality.asDict()
        return allArgs


class DetectPolicy(BaseDetectPolicy):
    """
    Class describes handler detect policy structure

    Attributes:
        estimatePeopleCount: people count estimation policy
        detectFace: whether to detect face from the image
        detectBody: whether to detect human body from the image
        estimateUpperBody: whether to estimate upper body attributes
        estimateLowerBody: whether to estimate lower body attributes
        estimateBasicBodyAttributes: whether to estimate basic body attributes
        estimateAccessories: whether to estimate accessories
    """

    def __init__(
        self,
        estimatePeopleCount: PeopleCount = None,
        detectFace: int = 1,
        detectBody: int = 0,
        estimateUpperBody: Optional[int] = None,
        estimateBasicBodyAttributes: Optional[int] = None,
        estimateAccessories: Optional[int] = None,
        estimateLowerBody: Optional[int] = None,
        **baseDetectPolicyKwargs,
    ):
        super().__init__(**baseDetectPolicyKwargs)
        self.estimatePeopleCount = estimatePeopleCount
        self.detectFace = detectFace
        self.detectBody = detectBody
        self.estimateUpperBody = estimateUpperBody
        self.estimateBasicBodyAttributes = estimateBasicBodyAttributes
        self.estimateAccessories = estimateAccessories
        self.estimateLowerBody = estimateLowerBody

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        result = {
            **super().asDict(),
            "detect_face": self.detectFace,
            "detect_body": self.detectBody,
        }
        if self.estimatePeopleCount is not None:
            result["estimate_people_count"] = self.estimatePeopleCount.asDict()
        if (
            self.estimateUpperBody is not None
            or self.estimateBasicBodyAttributes is not None
            or self.estimateAccessories is not None
            or self.estimateLowerBody is not None
        ):
            result["body_attributes"] = {}
            if self.estimateBasicBodyAttributes is not None:
                result["body_attributes"]["estimate_basic_attributes"] = self.estimateBasicBodyAttributes
            if self.estimateUpperBody is not None:
                result["body_attributes"]["estimate_upper_body"] = self.estimateUpperBody
            if self.estimateAccessories is not None:
                result["body_attributes"]["estimate_accessories"] = self.estimateAccessories
            if self.estimateLowerBody is not None:
                result["body_attributes"]["estimate_lower_body"] = self.estimateLowerBody
        return result


class BaseExtractPolicy:
    """
    Class describes extract policy base structure

    Attributes:
        extractBasicAttributes: flag that tells whether to extract face basic attributes (gender, age, ethnicity)
        fdScoreThreshold: face descriptor quality score threshold. All the attributes with quality score below the
                        threshold will be ignored (and not stored in the DB). The function will proceed as usual with
                        all the remaining descriptors (if left).
    """

    def __init__(self, extractBasicAttributes: int = 0, fdScoreThreshold: float = 0.0):
        self.extractBasicAttributes = extractBasicAttributes
        self.fdScoreThreshold = fdScoreThreshold

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        return {"extract_basic_attributes": self.extractBasicAttributes, "fd_score_threshold": self.fdScoreThreshold}


class ExtractPolicy(BaseExtractPolicy):
    """
    Class describes handler extract policy structure

    Attributes:
        extractFaceDescriptor: flag that tells whether to extract face descriptor(s)
        extractBodyDescriptor: flag that tells whether to extract human body descriptor(s)
    """

    def __init__(self, extractFaceDescriptor: int = 1, extractBodyDescriptor: int = 0, **baseExtractPolicyKwargs):
        super().__init__(**baseExtractPolicyKwargs)
        self.extractFaceDescriptor = extractFaceDescriptor
        self.extractBodyDescriptor = extractBodyDescriptor

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        return {
            **super().asDict(),
            "extract_face_descriptor": self.extractFaceDescriptor,
            "extract_body_descriptor": self.extractBodyDescriptor,
        }


@dataclass
class MatchPolicyCandidatesGeoFilter:
    """
    Class that describes matching candidates' geo filter.

    Attributes:
        originLongitude: origin longitude
        originLatitude: origin latitude
        longitudeDelta: longitude delta
        latitudeDelta: latitude delta
    """

    originLongitude: Real
    originLatitude: Real
    longitudeDelta: Optional[Real] = None
    latitudeDelta: Optional[Real] = None

    def asDict(self) -> Union[Dict[str, Real], None]:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        result = {
            "origin_longitude": self.originLongitude,
            "origin_latitude": self.originLatitude,
            "longitude_delta": self.longitudeDelta,
            "latitude_delta": self.latitudeDelta,
        }
        result = {key: value for key, value in result.items() if value is not None}
        return result or None


@dataclass
class MatchPolicyCandidates:
    """
    Class that describes matching candidates.

    Attributes:
        origin: object origin ("events" or "faces")
        accountId: object account id
        faceIds: object face ids
        externalIds: object external ids
        userData: object user data
        createTimeGte: object create time lower including boundary
        createTimeLt: object create time upper excluding boundary

        faceIdGte: face id lower including boundary
        faceIdLt: face id upper excluding boundary
        listId: face list id

        endTimeGte: object end time lower including boundary
        endTimeLt: object end time upper excluding boundary
        eventIds: event ids
        eventIdGte: event id lower including boundary
        eventIdLt: event id upper excluding boundary
        handlerIds: event handler ids
        topMatchingCandidatesLabel: event top matching candidates label
        topSimilarObjectIds: event top similar object ids
        topSimilarExternalIds: event top similar external ids
        topSimilarObjectSimilarityGte: event top similar object similarity lower including boundary
        topSimilarObjectSimilarityLt: event top similar object similarity upper excluding boundary
        ageGte: event age lower including boundary
        ageLt: event age upper excluding boundary
        gender: event gender
        emotions: event emotions
        masks: event masks
        liveness: event liveness states
        ethnicGroups: event ethnic groups
        sources: event sources
        streamIds: stream ids
        tags: event tags
        cities: event cities
        areas: event areas
        districts: event districts
        streets: event streets
        houseNumbers: event house numbers
        geoPosition: event geo position
        trackIds: event trackIds
        apparentAgeLt: apparent age upper excluding boundary
        apparentAgeGte: apparent age upper including boundary
        apparentGender: apparent gender estimation values (0 - female, 1 - male, 2 - undefined)
        headwearStates: headwear states (0 - absent, 1 - present, 2 - undefined)
        sleeveLengths: sleeve lengths
        upperClothingColors: upper clothing color set
        backpackStates: backpack states (0 - absent, 1 - present, 2 - undefined)
        meta: event meta filters
    """

    # faces or events:
    origin: str  # "events" or "faces"
    accountId: Optional[str] = MISSING
    faceIds: Optional[List[str]] = MISSING
    externalIds: Optional[List[str]] = MISSING
    userData: Optional[str] = MISSING
    createTimeGte: Optional[str] = MISSING
    createTimeLt: Optional[str] = MISSING
    # faces:
    faceIdGte: Optional[str] = MISSING
    faceIdLt: Optional[str] = MISSING
    listId: Optional[str] = MISSING
    # events:
    endTimeGte: Optional[str] = MISSING
    endTimeLt: Optional[str] = MISSING
    eventIds: Optional[List[str]] = MISSING
    eventIdGte: Optional[str] = MISSING
    eventIdLt: Optional[str] = MISSING
    handlerIds: Optional[List[str]] = MISSING
    topMatchingCandidatesLabel: Optional[str] = MISSING
    topSimilarObjectIds: Optional[List[str]] = MISSING
    topSimilarExternalIds: Optional[List[str]] = MISSING
    topSimilarObjectSimilarityGte: Optional[Real] = MISSING
    topSimilarObjectSimilarityLt: Optional[Real] = MISSING
    ageGte: Optional[int] = MISSING
    ageLt: Optional[int] = MISSING
    gender: Optional[int] = MISSING
    emotions: Optional[List[Union[int, None]]] = MISSING
    masks: Optional[List[Union[int, None]]] = MISSING
    liveness: Optional[List[Union[int, None]]] = MISSING
    ethnicGroups: Optional[List[Union[int, None]]] = MISSING
    sources: Optional[List[str]] = MISSING
    streamIds: Optional[List[str]] = MISSING
    tags: Optional[List[str]] = MISSING
    cities: Optional[List[Union[str, None]]] = MISSING
    areas: Optional[List[Union[str, None]]] = MISSING
    districts: Optional[List[Union[str, None]]] = MISSING
    streets: Optional[List[Union[str, None]]] = MISSING
    houseNumbers: Optional[List[Union[str, None]]] = MISSING
    geoPosition: Optional[Union[MatchPolicyCandidatesGeoFilter, dict, None]] = MISSING
    trackIds: Optional[List[Union[str, None]]] = MISSING
    apparentAgeLt: Optional[int] = MISSING
    apparentAgeGte: Optional[int] = MISSING
    apparentGender: Optional[List[Literal[0, 1, 2, None]]] = MISSING
    headwearStates: Optional[List[Literal[0, 1, 2, None]]] = MISSING
    sleeveLengths: Optional[List[Literal["short", "long", "undefined", None]]] = MISSING
    upperClothingColors: Optional[List[Union[CLOTHING_COLOR, None]]] = MISSING
    backpackStates: Optional[List[Literal[0, 1, 2]]] = MISSING
    lowerGarmentTypes: Optional[List[Literal["undefined", "trousers", "shorts", "skirt", None]]] = MISSING
    shoesApparentColors: Optional[List[Union[PRIMITIVE_CLOTHING_COLOR, None]]] = MISSING
    headwearApparentColors: Optional[List[Union[PRIMITIVE_CLOTHING_COLOR, None]]] = MISSING
    lowerGarmentColors: Optional[List[Union[CLOTHING_COLOR, None]]] = MISSING
    meta: Optional[dict] = MISSING

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure
        Returns:
            dict with values of this data structure
        """
        geoPosition = (
            self.geoPosition.asDict()
            if self.geoPosition is not MISSING and self.geoPosition is not None
            else self.geoPosition
        )
        result = {
            "origin": self.origin,
            "account_id": self.accountId,
            "face_ids": self.faceIds,
            "external_ids": self.externalIds,
            "user_data": self.userData,
            "create_time__gte": self.createTimeGte,
            "create_time__lt": self.createTimeLt,
            "end_time__gte": self.endTimeGte,
            "end_time__lt": self.endTimeLt,
            "face_id__gte": self.faceIdGte,
            "face_id__lt": self.faceIdLt,
            "list_id": self.listId,
            "event_ids": self.eventIds,
            "event_id__gte": self.eventIdGte,
            "event_id__lt": self.eventIdLt,
            "handler_ids": self.handlerIds,
            "top_matching_candidates_label": self.topMatchingCandidatesLabel,
            "top_similar_object_ids": self.topSimilarObjectIds,
            "top_similar_external_ids": self.topSimilarExternalIds,
            "top_similar_object_similarity__gte": self.topSimilarObjectSimilarityGte,
            "top_similar_object_similarity__lt": self.topSimilarObjectSimilarityLt,
            "age__gte": self.ageGte,
            "age__lt": self.ageLt,
            "gender": self.gender,
            "emotions": self.emotions,
            "masks": self.masks,
            "liveness": self.liveness,
            "ethnic_groups": self.ethnicGroups,
            "sources": self.sources,
            "stream_ids": self.streamIds,
            "tags": self.tags,
            "cities": self.cities,
            "areas": self.areas,
            "districts": self.districts,
            "streets": self.streets,
            "house_numbers": self.houseNumbers,
            "geo_position": geoPosition,
            "track_ids": self.trackIds,
            "apparent_age__lt": self.apparentAgeLt,
            "apparent_age__gte": self.apparentAgeGte,
            "apparent_gender": self.apparentGender,
            "headwear_states": self.headwearStates,
            "sleeve_lengths": self.sleeveLengths,
            "upper_clothing_colors": self.upperClothingColors,
            "backpack_states": self.backpackStates,
            "lower_garment_colors": self.lowerGarmentColors,
            "lower_garment_types": self.lowerGarmentTypes,
            "headwear_apparent_colors": self.headwearApparentColors,
            "shoes_apparent_colors": self.shoesApparentColors,
            "meta": self.meta,
        }
        result = {key: value for key, value in result.items() if value is not MISSING}
        return result


class MatchPolicy:
    """
    Class that describes matching conditions (list + some of basic attributes).
    Every event that satisfies the filters for the list will be matched by it

    Attributes:
        label: matching label (max length - 36)
        candidates: matching candidates
        descriptorType: candidate descriptor type
        gender: Gender. 1 - male, 0 - female.
        ethnicities: Array of integer (ethnicities_filter) (Items Enum:1 2 3 4)
        ageLt: [ 0 .. 100 ] - upper bound for reference age
        ageGte: [ 0 .. 100 ] - lower including bound for reference age
        limit: matching limit
        threshold: matching threshold
        targets: matching targets
        meta: event meta data
    """

    def __init__(
        self,
        candidates: MatchPolicyCandidates,
        label: Optional[str] = None,
        descriptorType: Optional[Literal["face", "body"]] = None,
        gender: Optional[int] = None,
        ethnicities: Optional[List[int]] = None,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        liveness: Optional[List[int]] = None,
        limit: Optional[int] = None,
        threshold: Optional[float] = None,
        targets: Optional[List[str]] = None,
    ):
        self.label = label
        self.candidates = candidates
        self.descriptorType = descriptorType
        self.gender = gender
        self.ethnicities = ethnicities
        self.ageLt = ageLt
        self.ageGte = ageGte
        self.limit = limit
        self.threshold = threshold
        self.targets = targets
        self.liveness = liveness

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"candidates": self.candidates.asDict(), "filters": {}}
        if self.label is not None:
            allArgs["label"] = self.label
        if self.descriptorType is not None:
            allArgs["descriptor"] = {"descriptor_type": self.descriptorType}
        if self.gender is not None:
            allArgs["filters"]["gender"] = self.gender
        if self.ethnicities is not None:
            allArgs["filters"]["ethnicities"] = self.ethnicities
        if self.ageLt is not None:
            allArgs["filters"]["age__lt"] = self.ageLt
        if self.ageGte is not None:
            allArgs["filters"]["age__gte"] = self.ageGte
        if self.liveness is not None:
            allArgs["filters"]["liveness"] = self.liveness
        if self.limit is not None:
            allArgs["limit"] = self.limit
        if self.threshold is not None:
            allArgs["threshold"] = self.threshold
        if self.targets is not None:
            allArgs["targets"] = self.targets
        return allArgs


class MatchFilter:
    """
    Describes a list for matching with optional similarity limitation

    Attributes:
        label: matching label - must equals one of matching policy matching labels
        similarityLte: [ 1 .. 0 ] Upper including bound for best match by the list
        similarityGte: [ 1 .. 0 ] Lower including bound for best match by the list
    """

    def __init__(self, label: str, similarityLte: Optional[float] = None, similarityGte: Optional[float] = None):
        self.label = label
        self.similarityLte = similarityLte
        self.similarityGte = similarityGte

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"label": self.label}
        if self.similarityLte is not None:
            allArgs["similarity__lte"] = self.similarityLte
        if self.similarityGte is not None:
            allArgs["similarity__gte"] = self.similarityGte
        return allArgs


class Filters:
    """
    Filters determine whether to create face or not

    Attributes:
        gender: Gender. 1 - male, 0 - female.
        ethnicities: Array of integer (ethnicities_filter) - Items Enum:1 2 3 4
                     ( AfricanAmerican - 1; Indian - 2; Asian - 3; Caucasian - 4)
        ageLt: [ 0 .. 100 ] - upper bound for reference age
        ageGte: [ 0 .. 100 ] - lower including bound for reference age
        liveness: Array of integer (filter by liveness state) - Items Enum: 0 1 2
        matchesFilters: a list of MatchFilter objects
    """

    def __init__(
        self,
        gender: Optional[int] = None,
        ethnicities: Optional[List[int]] = None,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        liveness: Optional[List[int]] = None,
        matchesFilters: Optional[List[MatchFilter]] = None,
    ):
        self.gender = gender
        self.ethnicities = ethnicities
        self.ageLt = ageLt
        self.ageGte = ageGte
        self.liveness = liveness
        self.matchesFilters = matchesFilters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {}
        if self.gender is not None:
            allArgs["gender"] = self.gender
        if self.ethnicities is not None:
            allArgs["ethnicities"] = self.ethnicities
        if self.ageLt is not None:
            allArgs["age__lt"] = self.ageLt
        if self.ageGte is not None:
            allArgs["age__gte"] = self.ageGte
        if self.liveness is not None:
            allArgs["liveness"] = self.liveness
        if self.matchesFilters is not None:
            allArgs["match"] = [matchesFilter.asDict() for matchesFilter in self.matchesFilters]
        return allArgs


class LinkToListPolicy:
    """
    Object with luna list for linking created faces

    Attributes:
        listId: luna list id
        filters: filters determine whether to link created face to the list
    """

    def __init__(self, listId: str, filters: Optional[Filters] = None):
        self.listId = listId
        self.filters = filters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"list_id": self.listId}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class FaceSamplePolicy:
    """
    Determines whether to create face samples

    Attributes:
        storeSample: whether to create sample from event
        filters: filters determine whether to create sample or not
    """

    def __init__(self, storeSample: int = 1, filters: Optional[Filters] = None):
        self.storeSample = storeSample
        self.filters = filters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"store_sample": self.storeSample}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class BodySamplePolicy(FaceSamplePolicy):
    """
    Determines whether to create body samples

    Attributes:
        storeSample: whether to create sample from event
        filters: filters determine whether to create sample or not
    """

    def __init__(self, storeSample: int = 1, filters: Optional[Filters] = None):
        super().__init__(storeSample=storeSample, filters=filters)


class ImageOriginPolicy:
    """
    Determines whether to create origin image

    Attributes:
        storeImage: whether to create origin image from event
        filters: filters determine whether to create origin image or not
        useExternalReferences: use external refereces for image origin or not
    """

    def __init__(self, storeImage: int = 0, filters: Optional[Filters] = None, useExternalReferences: int = 0):
        self.storeImage = storeImage
        self.filters = filters
        self.useExternalReferences = useExternalReferences

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"store_image": self.storeImage, "use_external_references": self.useExternalReferences}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class AttributePolicy:
    """
    Determines whether to create attributes

    Attributes:
        storeAttribute: whether to create attribute from event
        filters: filters determine whether to create attribute or not
        ttl: attribute storing time
    """

    def __init__(self, storeAttribute: int = 0, filters: Optional[Filters] = None, ttl: Optional[int] = None):
        self.storeAttribute = storeAttribute
        self.filters = filters
        self.ttl = ttl

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"store_attribute": self.storeAttribute}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        if self.ttl is not None:
            allArgs["ttl"] = self.ttl
        return allArgs


class FacePolicy:
    """
    Determines whether to create faces and link them to lists

    Attributes:
        storeFace: whether to create faces from events
        filters: filters determine whether to create face or not
        setSampleAsAvatar: whether to set a sample as avatar to created faces
        linkToListsPolicy: a set of luna lists for linking created faces
    """

    def __init__(
        self,
        storeFace: int = 0,
        filters: Optional[Filters] = None,
        setSampleAsAvatar: int = 1,
        linkToListsPolicy: Optional[List[LinkToListPolicy]] = None,
    ):
        self.storeFace = storeFace
        self.setSampleAsAvatar = setSampleAsAvatar
        self.filters = filters
        self.linkToListsPolicy = linkToListsPolicy

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"store_face": self.storeFace, "set_sample_as_avatar": self.setSampleAsAvatar}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        if self.linkToListsPolicy is not None:
            linkToListPolicyList = [linkToListPolicy.asDict() for linkToListPolicy in self.linkToListsPolicy]
            allArgs["link_to_lists_policy"] = linkToListPolicyList
        return allArgs


class EventPolicy:
    """
    Determines whether to create events

    Attributes:
        storeEvent: whether to create events
        filters: filters determine whether to create event or not
        waitSaving: whether to wait events saving (response will be received only after events will be saved)
    """

    def __init__(self, storeEvent: int = 1, filters: Optional[Filters] = None, waitSaving: Optional[int] = None):
        self.storeEvent = storeEvent
        self.filters = filters
        self.waitSaving = waitSaving

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"store_event": self.storeEvent}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        if self.waitSaving is not None:
            allArgs["wait_saving"] = self.waitSaving
        return allArgs


class NotificationPolicy:
    """
    Determines whether to send notifications

    Attributes:
        sendNotification: whether to send notification
        filters: filters determine whether to send notification or not
    """

    def __init__(self, sendNotification: int = 1, filters: Optional[Filters] = None):
        self.sendNotification = sendNotification
        self.filters = filters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"send_notification": self.sendNotification}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class ConditionalTagPolicy:
    """
    Determines whether to add a tag to event

    Attributes:
        tag: conditional tag
        filters: filters determine whether to link created face to the list
    """

    def __init__(self, tag: str, filters: Optional[Filters] = None):
        self.tag = tag
        self.filters = filters

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {"tag": self.tag}
        if self.filters is not None:
            allArgs["filters"] = self.filters.asDict()
        return allArgs


class StoragePolicy:
    """
    Object that determines a list of rules for luna objects creation. It can keep parameters
    for storing samples, attributes, faces, events.

    Attributes:
        faceSamplePolicy: policy determines params for creating face samples
        bodySamplePolicy: policy determines params for creating body samples
        imageOriginPolicy: policy determines params for creating origin images
        attributePolicy: policy determines params for creating attributes
        facePolicy: policy determines params for creating faces and link them to lists
        eventPolicy: policy determines params for creating events
        notificationPolicy: policy determines params for creating notifications
    """

    def __init__(
        self,
        faceSamplePolicy: Optional[FaceSamplePolicy] = None,
        bodySamplePolicy: Optional[BodySamplePolicy] = None,
        imageOriginPolicy: Optional[ImageOriginPolicy] = None,
        attributePolicy: Optional[AttributePolicy] = None,
        facePolicy: Optional[FacePolicy] = None,
        eventPolicy: Optional[EventPolicy] = None,
        notificationPolicy: Optional[NotificationPolicy] = None,
    ):
        self.faceSamplePolicy = faceSamplePolicy
        self.bodySamplePolicy = bodySamplePolicy
        self.imageOriginPolicy = imageOriginPolicy
        self.attributePolicy = attributePolicy
        self.facePolicy = facePolicy
        self.eventPolicy = eventPolicy
        self.notificationPolicy = notificationPolicy

    def asDict(self):
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {}
        if self.faceSamplePolicy is not None:
            allArgs["face_sample_policy"] = self.faceSamplePolicy.asDict()
        if self.bodySamplePolicy is not None:
            allArgs["body_sample_policy"] = self.bodySamplePolicy.asDict()
        if self.imageOriginPolicy is not None:
            allArgs["image_origin_policy"] = self.imageOriginPolicy.asDict()
        if self.attributePolicy is not None:
            allArgs["attribute_policy"] = self.attributePolicy.asDict()
        if self.facePolicy is not None:
            allArgs["face_policy"] = self.facePolicy.asDict()
        if self.eventPolicy is not None:
            allArgs["event_policy"] = self.eventPolicy.asDict()
        if self.notificationPolicy is not None:
            allArgs["notification_policy"] = self.notificationPolicy.asDict()
        return allArgs


class Policies:
    """
    Object that determines a list of rules for processing input images. It can keep parameters
    for detecting faces, extracting attributes. You can set rules for matching attributes and creating faces from them.

    Attributes:
        detectPolicy: policy determines params for face detecting.
        extractPolicy: policy determines params for attributes extracting.
        matchPolicy: list of luna lists with filters for matching. Every event that satisfies the filters
                     for the list will be matched by it
        storagePolicy: policy for objects creation
        conditionalTagsPolicy: a set of conditional tags for event
    """

    def __init__(
        self,
        detectPolicy: Optional[DetectPolicy] = None,
        extractPolicy: Optional[ExtractPolicy] = None,
        matchPolicy: Optional[List[MatchPolicy]] = None,
        storagePolicy: Optional[StoragePolicy] = None,
        conditionalTagsPolicy: Optional[List[ConditionalTagPolicy]] = None,
    ):
        self.detectPolicy = detectPolicy
        self.extractPolicy = extractPolicy
        self.matchPolicy = matchPolicy
        self.storagePolicy = storagePolicy
        self.conditionalTagsPolicy = conditionalTagsPolicy

    def asDict(self) -> dict:
        """
        Creates a dict with all existing values of this data structure

        Returns:
            dict with values of this data structure
        """
        allArgs = {}
        if self.detectPolicy is not None:
            allArgs["detect_policy"] = self.detectPolicy.asDict()

        if self.extractPolicy is not None:
            allArgs["extract_policy"] = self.extractPolicy.asDict()

        if self.matchPolicy is not None:
            allArgs["match_policy"] = [element.asDict() for element in self.matchPolicy]

        if self.storagePolicy is not None:
            allArgs["storage_policy"] = self.storagePolicy.asDict()

        if self.conditionalTagsPolicy is not None:
            tagPolicyList = [policy.asDict() for policy in self.conditionalTagsPolicy]
            allArgs["conditional_tags_policy"] = tagPolicyList
        return allArgs


class VerifierDetectPolicy(BaseDetectPolicy):
    """Class describes verifier detect policy structure."""

    pass  # pylint: disable=W0107


class VerifierExtractPolicy(BaseExtractPolicy):
    """Class describes verifier extract policy structure."""

    pass  # pylint: disable=W0107


class VerifierFaceSamplePolicy:
    """
    Determines whether to create face samples

    Attributes:
        storeSample: whether to create sample from event

    """

    def __init__(self, storeSample: int = 0):
        self.storeSample = storeSample

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        return {"store_sample": self.storeSample}


class VerifierAttributePolicy:
    """
    Determines whether to create attributes

    Attributes:
        storeAttribute: whether to create attribute

    """

    def __init__(self, storeAttribute: int = 0):
        self.storeAttribute = storeAttribute

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        return {"store_attribute": self.storeAttribute}


class VerifierStoragePolicy:
    """
    Verifier object's storing policy

    Attributes:
        faceSamplePolicy: face sample storing policiy
        attributePolicy: attribute storing policy

    """

    def __init__(
        self,
        faceSamplePolicy: Optional[VerifierFaceSamplePolicy] = None,
        attributePolicy: Optional[VerifierAttributePolicy] = None,
    ):
        self.faceSamplePolicy = faceSamplePolicy
        self.attributePolicy = attributePolicy

    def asDict(self):
        """Creates dictionary representation."""
        allArgs = {}
        if self.faceSamplePolicy is not None:
            allArgs["face_sample_policy"] = self.faceSamplePolicy.asDict()
        if self.attributePolicy is not None:
            allArgs["attribute_policy"] = self.attributePolicy.asDict()
        return allArgs


class VerifierPolicies:
    """
    Object that determines a list of rules for processing & verification of input images. It can keep parameters
    for detecting faces, extracting attributes. You can set rules for matching attributes and creating faces from them.

    Attributes:
        verificationThreshold: similarity threshold for verification
        detectPolicy: policy determines params for face detecting.
        extractPolicy: policy determines params for attributes extracting.
        storagePolicy: policy determines params for objects storing

    """

    def __init__(
        self,
        verificationThreshold: Optional[float] = None,
        detectPolicy: Optional[VerifierDetectPolicy] = None,
        extractPolicy: Optional[VerifierExtractPolicy] = None,
        storagePolicy: Optional[VerifierStoragePolicy] = None,
    ):
        self.verificationThreshold = verificationThreshold
        self.detectPolicy = detectPolicy
        self.extractPolicy = extractPolicy
        self.storagePolicy = storagePolicy

    def asDict(self) -> dict:
        """Creates dictionary representation."""
        allArgs = {}
        if self.verificationThreshold is not None:
            allArgs["verification_threshold"] = self.verificationThreshold
        if self.detectPolicy is not None:
            allArgs["detect_policy"] = self.detectPolicy.asDict()
        if self.extractPolicy is not None:
            allArgs["extract_policy"] = self.extractPolicy.asDict()
        if self.storagePolicy is not None:
            allArgs["storage_policy"] = self.storagePolicy.asDict()
        return allArgs


class EstimatorHandler:
    """
    Handler object for estimator task (the new handler data)

    Args:
        policies: input image processing policies
        description: handler description
    """

    def __init__(self, policies: Policies, description: Optional[str] = None):
        self.policies = policies
        self.description = description

    def asDict(self) -> dict:
        """Prepare dict representation."""
        handler = {"policies": self.policies.asDict()}
        if self.description is not None:
            handler["description"] = self.description
        return handler


class EstimatorHandlerExtant:
    """
    Handler object for estimator task (the extant handler data)

    Args:
        handlerId: handler id
        policies: input image processing policies (for dynamic handler)
    """

    def __init__(self, handlerId: str, policies: Optional[Policies] = None):
        self.handlerId = handlerId
        self.policies = policies

    def asDict(self) -> dict:
        """Prepare dict representation."""
        handler = {"handler_id": self.handlerId}
        if self.policies is not None:
            handler["policies"] = self.policies.asDict()
        return handler


class GeoBbox:
    """
    Geo position bbox

    Attributes:
        originLongitude: bbox origin longitude, in degrees
        originLatitude: bbox origin latitude, in degrees
        longitudeDelta: bbox longitude delta, in degrees
        latitudeDelta: bbox latitude delta, in degrees
    """

    def __init__(
        self,
        originLongitude: float,
        originLatitude: float,
        longitudeDelta: Optional[float] = None,
        latitudeDelta: Optional[float] = None,
    ):
        self.originLongitude = originLongitude
        self.originLatitude = originLatitude
        self.longitudeDelta = longitudeDelta
        self.latitudeDelta = latitudeDelta

    def __repr__(self):
        return self.buildQueryObject()

    def buildQueryObject(self) -> str:
        """Build query object."""
        data = f"origin_longitude,{self.originLongitude},origin_latitude,{self.originLatitude}"
        if self.longitudeDelta is not None:
            data += f",longitude_delta,{self.longitudeDelta}"
        if self.latitudeDelta is not None:
            data += f",latitude_delta,{self.latitudeDelta}"
        return data
