"""
Module implements http response
"""
import asyncio
import uuid
from contextlib import asynccontextmanager
from time import perf_counter
from typing import AsyncGenerator, Mapping, Optional, Union

import aiohttp
import chardet
import msgpack
import ujson
from aiohttp import DummyCookieJar
from aiohttp.client_reqrep import json_re

from ..common.aiohttp_exception import guessAiohttpError
from ..common.luna_request import LunaRequest
from .exceptions import LunaApiException
from .logger import logger
from .utils import DNSResolver


class LunaResponse:
    """
    The producer and container class for aiohttp response.

    Attributes:

        rawResponse: aiohttp client response
        request: sent request
        raiseError: whether to raise LunaApiException in case of failure
        echo: log request and response message or not

        _responseBody: rawResponse body
        _customBody: custom response body if _responseBody has wrong format
        _json: loaded json
        _status: processing status code
        _headers: rawResponse headers
        _session: session
    """

    __slots__ = (
        "rawResponse",
        "request",
        "raiseError",
        "_customBody",
        "_responseBody",
        "_json",
        "_status",
        "_headers",
        "_session",
        "echo",
    )

    def __init__(
        self,
        request: LunaRequest,
        session: Optional[aiohttp.ClientSession] = None,
        raiseError: bool = False,
        echo: bool = False,
    ):
        self.rawResponse: Optional[aiohttp.ClientResponse] = None
        self._customBody: Optional[str] = None
        self._responseBody: Optional[bytes] = None
        self._json: Union[dict, list, int, None] = None
        self._status: Optional[int] = None
        self._headers: dict = {}

        self.request = request
        self._session = session
        self.raiseError = raiseError
        self.echo = echo

    @asynccontextmanager
    async def sendRequest(self) -> AsyncGenerator["LunaResponse", None]:
        """
        Send request and yield prepared response (both success and failure).
        In case of http client error (connection failure, timeout, etc) plant fake response.

        Yields:
            LunaResponse instance
        """
        if self.echo:
            urid = str(uuid.uuid4())[:6]
            request = self.request
            requestBody = "<not load>"
            if self.request.json:
                requestBody = str(request.json)[:200]

            logger.info(
                f"Begin request. id: {urid}, method: {request.method}, url: {request.url}, headers: {request.headers},"
                f"body: {requestBody}"
            )
            startTime = perf_counter()

        if self._session is None:
            session = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(resolver=DNSResolver()), cookie_jar=DummyCookieJar()
            )
        else:
            session = self._session

        try:
            async with session.request(**self.request) as response:
                self._status, self.rawResponse = response.status, response

                if self.raiseError and not self.success:
                    # force reading the response to provide details for LunaApiException
                    await self.read()
                    self.raiseForStatus()

                yield self
        except (asyncio.TimeoutError, aiohttp.ClientError) as exc:
            # Plant fake response
            if isinstance(exc, asyncio.TimeoutError):
                self._status = 504
            elif isinstance(exc, aiohttp.ClientError):
                self._status = 500

            self._customBody = ujson.dumps(guessAiohttpError(exc, self.request))
            self._headers["Content-Type"] = f"application/json; charset={self.charset}"

            if self.raiseError:
                self.raiseForStatus()

            yield self

        finally:
            if self.echo:
                requestTime = perf_counter() - startTime
                responseBody = self.text[:200]
                logger.info(
                    f"End request. id: {urid}, execution_time: {requestTime:.5f}, "
                    f"status_code: {self._status}, headers: {self.headers}, body: {responseBody}"
                )

            if self._session is None:
                await session.close()

    @property
    def requestId(self) -> Union[None, str]:
        """
        Returns:
            response 'Luna-Request-Id' header if exists
            request 'Luna-Request-Id' if response 'Luna-Request-Id' header does not exist and
                request 'Luna-Request-Id' header exists
            None otherwise
        """
        return self.headers.get("Luna-Request-Id") or self.request.headers.get("Luna-Request-Id")

    @property
    def headers(self) -> Mapping[str, str]:
        """
        Response headers
        Returns:
            headers of the response
            {} if no response
        """
        if self.rawResponse is not None:
            return self.rawResponse.headers
        return self._headers

    @property
    def contentType(self):
        """
        Get headers Content-Type
        Returns:
            optional Content-Type
        """
        return self.headers.get("Content-Type")

    @property
    def statusCode(self) -> int:
        """
        Get status code from response
        Returns:
            status code
        """
        return self._status

    @property
    def body(self) -> Union[bytes, str, None]:
        """
        Choose a body to return.

        Returns:
            'None' if request was not fetched
            self._customBody if set
            self._responseBody otherwise
        """
        if self._customBody is not None:
            return self._customBody
        return self._responseBody

    @property
    def charset(self) -> Union[str, None]:
        """
        Wrapper around aiohttp.Response.

        In case customBody is set returns 'utf-8'.

        Returns:
            charset of body if specified
        """
        if self._customBody is not None:
            return "utf-8"

        return self.rawResponse.charset

    @property
    def text(self) -> str:
        """
        Convert body to string.

        Encoding will be guessed using ``chardet``.

        Returns:
             decode body
        """
        #: if we setted body
        if isinstance(self.body, str):
            return self.body

        if self.body is None:
            return "<not loaded>"

        charset = self.charset
        if charset is None:
            # Assume that the body is encoded in utf-8
            charset = "utf-8"

        try:
            content = self.body.decode(charset)
        except (LookupError, UnicodeDecodeError, TypeError):
            # Try to decode with guessed encoding
            encoding = chardet.detect(self.body)["encoding"]
            try:
                content = self.body.decode(encoding)
            except (LookupError, UnicodeDecodeError, TypeError):
                # A LookupError is raised if the encoding was not found which could
                # indicate a misspelling or similar mistake.
                #
                # A TypeError can be raised if encoding is None
                #
                # So we try blindly encoding.
                content = str(self.body, errors="replace")
        return content

    @property
    def json(self) -> Union[dict, list, int]:
        """
        Deserialize body to a Python object.
        Body contains either JSON string, either MessagePack bytes.

        Returns:
            A Python object
        """
        if self._json:
            return self._json

        errorMsg = None

        if self.contentType == "application/msgpack":
            try:
                self._json = msgpack.unpackb(self.body, raw=False)
            except (ValueError, msgpack.UnpackException) as exc:
                errorMsg = f"Failed to unpack MessagePack: '{exc}'"
        else:
            try:
                self._json = ujson.loads(self.text)
            except ValueError:
                if json_re.match(self.contentType) and self.request.method.upper() != "HEAD":
                    # could be json
                    errorMsg = f"Not json in response: '{self.text}'"

        if errorMsg is not None:
            logger.error(errorMsg)
            self._json = {"error_code": 24, "desc": "HTTP client error", "detail": errorMsg}
            self._status = 500
            self._customBody = ujson.dumps(self._json)

        return self._json

    @property
    def success(self) -> bool:
        """
        Property for checking whther the reqiest was successful or not.

        Returns:
            True if statusCode is less than 400, False if not
        """
        return self.statusCode < 400

    def __repr__(self):
        responseBody = "<not load>"
        if self.contentType is not None and (
            json_re.match(self.contentType)
            or self.contentType.startswith("text/json")
            or self.contentType.startswith("text/plain")
        ):
            responseBody = self.text[:200]
        return f"<LunaResponse> status code: {self.statusCode}, headers: {self.headers}, body: {responseBody}"

    async def read(self) -> None:
        """
        Method for fetching body chunks.
        Also decide about the correctness of the incoming response body.
        """
        if self._customBody is not None:
            # no need to proceed in case response was planted due to timeout or client error
            return

        self._responseBody = await self.rawResponse.read()

        if self.statusCode >= 400:
            if self.request.method.upper() != "HEAD" and (
                self.contentType is None or not json_re.match(self.contentType)
            ):
                errorMsg = (
                    f"{self.rawResponse.method}, {self.rawResponse.url}, {self.requestId}, "
                    f"Attempt to load error with unexpected Content-Type, status code: {self.statusCode}"
                )
                self._status = 500
                self._json = {"error_code": 24, "desc": "HTTP client error", "detail": errorMsg}
                self._customBody = ujson.dumps(self._json)
                logger.error(errorMsg)
                try:
                    logger.error(self.text)
                except Exception as exc:  # pylint: disable=W0703
                    logger.error(
                        f"Failed to log a response text: {exc!r}, uri: {self.request.url}, "
                        f"method: {self.request.method}"
                    )

        elif self.statusCode == 204:
            if self._responseBody:
                logger.warning(
                    f"Resource {self.rawResponse.url} return status code 204 with not empty body on"
                    f"method {self.rawResponse.method} "
                )

    def raiseForStatus(self):
        """
        :class:`~.LunaApiException` generator.

        If the request fails, an exception is thrown.

        Raises:
            LunaApiException: if *self.success* is not success
        """
        if self.success:
            return self

        msg = "Not expected code from service: {} {} {}, error: {}".format(
            self.statusCode, self.request.method, self.request.url, self.text
        )

        raise LunaApiException(msg, self)

    async def iter(self, chunkSize=100 * 1024 * 1024) -> AsyncGenerator[bytes, None]:
        """
        Yield chunks from the stream.

        Args:
            chunkSize: maximum size of the chunk in bytes
        """
        async for chunk in self.rawResponse.content.iter_chunked(chunkSize):
            yield chunk
