"""
Luna-license client module
"""
from numbers import Real
from typing import Awaitable, List, Literal, Optional, Union

import aiohttp

from ..common.base_api import BaseHTTPAPI
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload

TARGETS = Literal[
    "expiration_time", "faces_limit", "liveness", "iso", "index", "people_counter", "body_attributes", "streams_limit"
]


class LicensesApi(BaseHTTPAPI):
    """
    Class for request to luna-licenses.

    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5120",
        api: Optional[int] = 1,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def getLicense(
        self,
        targets: Optional[List[TARGETS]] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get current license data.

        Args:
            targets: group of features to return

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with license info.

        """
        queries = {
            "targets": targets,
        }
        return self.makeRequest(f"{self.baseUri}/license", "GET", queryParams=queries, **kwargs)

    def consumeExecutions(self, liveness: int, **kwargs):
        """
        Consume executions for specified features
        Args:
            liveness: count of executions for liveness to consume
        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return json with license info.
        """
        payload = RequestPayload.buildJson(body=dict(liveness=dict(execution_count=liveness)))
        return self.makeRequest(f"{self.baseUri}/license/executions", "POST", body=payload, **kwargs)
