"""
Module implements sessions to services.
"""
from abc import ABC, abstractmethod
from dataclasses import asdict
from typing import Optional, Type, Union

import aiohttp

from .common.base_api import BaseAPI, BaseHTTPAPI
from .settings import HTTPSettings


class AbstractServiceSession(ABC):
    """Session to a service. Can contain some status."""

    @abstractmethod
    def getClient(self, lunaRequestId: Optional[str] = None) -> BaseAPI:
        """
        Allows to get a service client with specified request id.

        Args:
            lunaRequestId: request id

        Returns:
            service client
        """


class HTTPSession(AbstractServiceSession):
    """
    Session for HTTP service.

    Attributes:
        RESTClient: service rest client
        settings: settings to unpack
        echo: log a request and response

        _aiohttpSession: client session lazy container
    """

    def __init__(
        self, RESTClient: Type[BaseHTTPAPI], settings: HTTPSettings, echo: bool = False
    ):  # pylint: disable=invalid-name
        self.RESTClient: Type[BaseHTTPAPI] = RESTClient  # pylint: disable=C0103
        self.settings: HTTPSettings = settings
        self.echo = echo

        self._aiohttpSession: Union[None, aiohttp.ClientSession] = None

    def getClient(self, lunaRequestId: Optional[str] = None) -> BaseHTTPAPI:
        """
        Get service client with specified lunaRequestId.

        Args:
            lunaRequestId: request id

        Returns:
            HTTP client
        """
        if self._aiohttpSession is None:
            client = self.RESTClient(**asdict(self.settings), lunaRequestId=lunaRequestId, echo=self.echo)
            self._aiohttpSession = client.session
        else:
            client = self.RESTClient(
                **asdict(self.settings), lunaRequestId=lunaRequestId, session=self._aiohttpSession, echo=self.echo
            )
        return client

    async def close(self):
        """
        Close current aiohttp session if necessary.
        """
        if self._aiohttpSession is not None:
            await self._aiohttpSession.close()
