""" Some structures for Streams Api. """
from dataclasses import asdict, dataclass
from typing import List, Optional, Union

from stringcase import snakecase


@dataclass
class Handler:
    """Bestshot/detection handler structure"""

    handlerId: str

    def asDict(self):
        """Creates dict representation."""
        return {"handler_id": self.handlerId}


@dataclass
class Authorization:
    """Stream event handler authorization"""

    token: Optional[str] = None
    accountId: Optional[str] = None

    def asDict(self):
        """Creates dict representation"""
        if self.token:
            return {"token": self.token}
        return {"account_id": self.accountId}


@dataclass
class EventHandler:
    """Stream event handler structure."""

    origin: str
    apiVersion: int
    bestshotHandler: Handler
    detectionHandler: Optional[Handler] = None
    frameStore: Optional[str] = None
    authorization: Optional[Authorization] = None

    def asDict(self):
        """Creates dict representation."""
        data = {
            "origin": self.origin,
            "api_version": self.apiVersion,
            "bestshot_handler": self.bestshotHandler.asDict(),
        }
        if self.detectionHandler:
            data["detection_handler"] = self.detectionHandler.asDict()
        if self.frameStore:
            data["frame_store"] = self.frameStore
        if self.authorization:
            data["authorization"] = self.authorization.asDict()
        return data


@dataclass
class GeoPosition:
    """Geo position class"""

    latitude: float
    longitude: float


@dataclass
class Location:
    """Location structure."""

    area: Optional[str] = None
    city: Optional[str] = None
    district: Optional[str] = None
    street: Optional[str] = None
    houseNumber: Optional[str] = None
    geoPosition: Optional[GeoPosition] = None

    def asDict(self):
        """Creates dict representation."""
        return {snakecase(field): value for field, value in asdict(self).items() if value is not None}


@dataclass
class Restart:
    """Restart configuration structure."""

    restart: Optional[bool] = False
    attemptCount: Optional[int] = 10
    delay: Optional[int] = 60

    def asDict(self):
        """Creates dict representation."""
        return {snakecase(field): int(value) for field, value in asdict(self).items() if value is not None}


@dataclass
class QueueFilters:
    """Streams queue filters"""

    streamIds: Optional[List[str]] = None
    names: Optional[List[str]] = None
    groupIds: Optional[List[str]] = None
    groupNames: Optional[List[str]] = None

    def asDict(self) -> dict:
        """Creates dict representation."""
        return {snakecase(field): value for field, value in asdict(self).items() if value is not None}


@dataclass
class StreamVideoInfo:
    """Stream feedback video info class."""

    width: int
    height: int
    frameRate: int
    bitRate: int
    gopSize: int
    startTime: str
    duration: Optional[float] = None
    progress: Optional[float] = None

    def asDict(self):
        """Creates dict representation."""
        return {snakecase(field): value for field, value in asdict(self).items() if value is not None}


@dataclass
class PreviewUrl:
    """Preview url class."""

    url: str


@dataclass
class Preview:
    """Streams preview info class."""

    live: Union[PreviewUrl, None]
    lastFrame: Union[PreviewUrl, None]

    def asDict(self):
        """Created dict representations."""
        return {snakecase(field): value for field, value in asdict(self).items() if value is not None}


@dataclass
class StreamFeedback:
    """Stream feedback class."""

    streamId: str
    status: str
    version: int
    error: Union[str, None]
    videoInfo: Union[StreamVideoInfo, None]
    preview: Union[Preview, None]

    def asDict(self):
        """Creates dict representation."""
        result = {snakecase(field): getattr(self, field) for field in ("streamId", "status", "version", "error")}
        result["video_info"] = self.videoInfo and self.videoInfo.asDict()
        if self.preview:
            result["preview"] = self.preview.asDict()
        return result
